local M = {}

local scenario = require("scenario/scenarios")
local helper = require("scenario/scenariohelper")
local im = ui_imgui
local ffi = require('ffi')

local players = {}
local playerNames = {"ad_player1", "ad_player2"}
local playerVehicleFullNames = {"", ""}
local trafficGroupName = "ad_traffic"
local trafficIDs = {}
local wpGroupName = "ad_wp"
local wpMiddle = 6
local wpFinal = 12
local switchMarker, finishMarker = nil, nil
local newStartPos, newStartRot = nil, nil
local eventState = "ready"
local timers = {}
local timeRanks = {110, 130, 220}
local playerControl = 1
local forceProgress = false
local forceState = 0
local forceDelta = 0
local win = true
local fullWidth = TorqueScript.eval('getDesktopResolution();') and tonumber(string.match(TorqueScript.eval('getDesktopResolution();'), "(%d+)%s")) or 1280
local fullHeight = nil

local windowOpen = im.BoolPtr(false)
local baseColor = im.ImVec4(1, 1, 1, 1)
local activeColor = im.ImVec4(1, 0.8627, 0.4784, 1)
local styles = {}

local text = {
	win = "Finish!",
	time1 = "Wow, that's a blazing fast time! You're on fire!",
	time2 = "Awesome driving; here, you've earned this gold medal!",
	time3 = "You did good out there.",
	sameCar = "Impressive, you chose the same car configuration for both parts of the run! You can really handle that thing anywhere, eh?",
	crash = "That's not good. Try again!",
	bad = "You didn't enter your second vehicle!",
	stupid = "Why in the world did you hit your other car?"
}
local group = {
	{"bluebuck"},
	{"burnside", "3a"},
	{"coupe", "type-l_A"},
	{"etki"},
	{"hopper"},
	{"legran"},
	{"moonhawk", "v8a_alt"},
	{"pickup"},
	{"pickup", "ext_4wd"},
	{"van"},
	{"van", "deliverytruck"},
	{"sunburst", "sport_DCT"}
}

local quantity = Engine.Platform.getCPUInfo().coresPhysical or 4
quantity = clamp(quantity - 2, 2, 3)
local damageLimit = 10000

local function forceWaypoints()
	forceState = forceState + 1
	local up = vec3(0, 0, 1)
	if forceState <= wpMiddle then
		local pos = vec3(scenetree.findObject(wpGroupName..forceState):getPosition()) + up * 5
		players[2].obj:setPositionRotation(pos.x, pos.y, pos.z, 0, 0, 0, 1)
	else
		local pos = newStartPos
		local rot = quatFromDir(vec3(0, -1, 0):rotated(quatFromDir(newStartRot, up)), up)
		players[2].obj:setPositionRotation(pos.x, pos.y, pos.z, rot.x, rot.y, rot.z, rot.w)
		forceProgress = false
	end
end

local function getVehicleFullName(name)
	local fullName = nil
	local str1, str2 = "", ""
	local obj = scenetree.findObject(name)
	local modelName = obj.jbeam
	local configName = string.match(obj.partConfig, "/([%w|_]+).pc")
	local json1 = jsonReadFile("vehicles/"..modelName.."/info.json")
	if json1 then
		if json1.Brand and json1.Name then
			str1 = json1.Brand.." "..json1.Name
		elseif json1.Name then
			str1 = json1.Name
		else
			str1 = modelName
		end
		local json2 = jsonReadFile("vehicles/"..modelName.."/info_"..configName..".json")
		if json2 then
			str2 = json2.Configuration and json2.Configuration or configName
		end
	end
	fullName = str1.." "..str2
	return fullName
end

local function getTextColor(int)
	return be:getPlayerVehicle(0).name == playerNames[int] and activeColor or baseColor
end

local function showExtraWindow()
	if not windowOpen[0] then return end
	im.SetNextWindowPos(im.ImVec2(fullWidth / 2 - 240, 100), im.Cond_Always)
	im.SetNextWindowSize(im.ImVec2(480, 120), im.Cond_Always)
	for k, v in pairs(styles) do
		im.PushStyleColor2(k, v)
	end
	if im.Begin("Go To Vehicle ", windowOpen, im.WindowFlags_NoResize) then
		im.SetWindowFontScale(1.4)
		if im.Button("Vehicle #1") then
			be:enterVehicle(0, scenetree.findObject(playerNames[1]))
        end
		im.SameLine()
		im.TextColored(getTextColor(1), playerVehicleFullNames[1])
		if im.Button("Vehicle #2") then
			be:enterVehicle(0, scenetree.findObject(playerNames[2]))
        end
		im.SameLine()
		im.TextColored(getTextColor(2), playerVehicleFullNames[2])
		im.End()
	end
	for k, v in pairs(styles) do
		im.PopStyleColor()
	end
end

local function onVehicleResetted(id)
	if id == be:getPlayerVehicle(0):getID() then
		getVehicleFullName(be:getPlayerVehicle(0).name)
	end
end

local function onScenarioLoaded()
	core_multiSpawn.setGroup("custom", group)
	core_multiSpawn.setupVehicles(quantity, false, "road", 60)
end

local function onScenarioUIReady()
	eventState = "screen"
	players = {}
	playerControl = 1
	timers = {switch = 0, fail = 0, p1 = 0, p2 = 0}
	switchMarker = scenetree.findObject("switchPoint")
	finishMarker = scenetree.findObject(wpGroupName..wpFinal)
	styles[10] = im.ImVec4(0.8, 0.3216, 0, 1)
	styles[11] = im.ImVec4(0.8, 0.3216, 0, 1)
	styles[21] = im.ImVec4(1, 0.4039, 0, 1)
	styles[22] = im.ImVec4(1, 0.5216, 0.2, 1)
	styles[23] = im.ImVec4(0.8, 0.3216, 0, 1)
	for i, v in ipairs(playerNames) do
		playerVehicleFullNames[i] = getVehicleFullName(v)
	end
	gameplay_traffic.deactivate()
end

local function onCountdownStarted()
	eventState = "ready"
	for i, v in ipairs(playerNames) do
		local a = {}
		a.obj = scenetree.findObject(v)
		a.id = a.obj:getID()
		table.insert(players, a)
	end
	newStartPos = vec3(players[2].obj:getPosition())
	newStartRot = vec3(players[2].obj:getDirectionVector())
	local others = core_multiSpawn.getOtherVehicles()
	local int = 1
	for i, v in ipairs(others) do
		if not string.find(scenetree.findObjectById(v).name, "player") then
			scenetree.findObjectById(v):setField("name", 0, trafficGroupName..int)
			int = int + 1
		end
	end
	win = true
	be:enterVehicle(0, scenetree.findObject(playerNames[1]))
	gameplay_traffic.activate()
	for i = 1, 6, 1 do
		local new = scenetree.findObject(trafficGroupName..i)
		if new then
			gameplay_traffic.insertTraffic(new:getID())
			table.insert(trafficIDs, new:getID())
		end
	end
	for i, v in ipairs(trafficIDs) do
		local value = math.random() * 0.5
		local vars = {spawnIn = value, spawnOut = value, persist = true}
		gameplay_traffic.setVehicleVars(vars, v)
	end
end

local function onRaceStart()
	switchMarker:setPosition(scenetree.findObject("switchZone"):getPosition())
	forceProgress = true
	eventState = "part1"
end

local function onRaceTick(tick)
	for i, v in ipairs(players) do
		local n = i == 1 and 2 or 1
		v.pos = vec3(v.obj:getPosition())
		v.damage = map.objects[v.id].damage
		v.vel = map.objects[v.id].vel:length()
		v.dist1 = v.pos:distance(vec3(switchMarker:getPosition()))
		v.dist2 = v.pos:distance(vec3(finishMarker:getPosition()))
		if timers.fail >= 4 then
			win = false
			scenario.finish({failed = text.crash})
		end
		if i == 1 then
			if v.dist2 <= 20 then
				win = false
				scenario.finish({failed = text.bad})
			end
			if eventState == "part1" and v.dist1 <= 8 then
				switchMarker:setPosition(vec3(0, 0, 2000):toPoint3F())
				playerControl = 2
				eventState = "switch"
			end
		elseif i == 2 then
			if v.dist1 <= 20 and v.damage >= damageLimit then
				win = false
				players[1].obj:queueLuaCommand("ai.setMode('disabled')")
				scenario.finish({failed = text.stupid})
			end
			if eventState == "newStart" and players[1].vel <= 1 then
				players[1].obj:queueLuaCommand("ai.setMode('disabled')")
				eventState = "part2"
			end
		end
		if i == playerControl then
			if v.damage >= damageLimit then
				if v.vel <= 4 then
					timers.fail = timers.fail + tick
				else
					timers.fail = 0
				end
			end
			timers["p"..i] = timers["p"..i] + tick
		end
	end
	if playerControl == 2 then
		if timers.switch >= 2 then
			if eventState == "newReady" then
				players[2].obj:queueLuaCommand("controller.setFreeze(0)")
				Engine.Audio.playOnce('AudioGui', 'event:UI_CountdownGo')
				helper.flashUiMessage("GO!", 2, true)
				for i, v in ipairs(trafficIDs) do
					local vars = {spawnIn = 0, spawnOut = 0, persist = true}
					gameplay_traffic.setVehicleVars(vars, v)
				end
				eventState = "newStart"
			end
		elseif timers.switch >= 1 then
			if eventState == "switch" then
				helper.flashUiMessage("Ready...", 0.8, true)
				eventState = "newReady"
			end
		elseif timers.switch == 0 then
			be:enterVehicle(0, scenetree.findObject(playerNames[2]))
			players[1].obj:queueLuaCommand("ai.setMode('random')")
			players[1].obj:queueLuaCommand("ai.setSpeedMode('limit')")
			players[1].obj:queueLuaCommand("ai.setSpeed(0)")
			players[2].obj:queueLuaCommand("controller.setFreeze(1)")
		end
		timers.switch = timers.switch + tick
	end
end

local function onRaceResult()
	if win then
		local finishTime = scenario.getScenario().timer
		local finishText = text.win
		local timeText = nil
		for i, v in ipairs(timeRanks) do
			if finishTime <= v then
				timeText = text["time"..i]
				break
			end
		end
		if timeText then
			finishText = finishText.." "..timeText
		end
		if players[1].obj.jbeam == players[2].obj.jbeam and players[1].obj.partConfig == players[2].obj.partConfig then
			finishText = finishText.." "..text.sameCar
		end
		scenario.finish({msg = finishText})
	end
end

local function onUpdate(dt)
	windowOpen[0] = (core_gamestate.state.state == "scenario" and eventState == "screen") and true or false
	showExtraWindow()
	if forceProgress then
		forceDelta = forceDelta + dt
		if forceDelta >= 0.1 then
			forceWaypoints()
			forceDelta = 0
		end
	end
end

M.onVehicleResetted = onVehicleResetted
M.onScenarioLoaded = onScenarioLoaded
M.onScenarioUIReady = onScenarioUIReady
M.onCountdownStarted = onCountdownStarted
M.onRaceStart = onRaceStart
M.onRaceTick = onRaceTick
M.onRaceResult = onRaceResult
M.onUpdate = onUpdate

return M